import unittest
import numpy as np
import os

import ufss

def calculate_spectrum_with_UF2(interaction_picture_shift):
    folder = os.path.join('fixtures','copolymer_P1')

    uf2 = ufss.DensityMatrices(os.path.join(folder,'open'),
                               conserve_memory=True,
                               detection_type='complex_polarization')

    uf2.interaction_picture_shift = interaction_picture_shift

    # defining the optical pulses in the RWA
    sigma = 1
    t = np.arange(-20,21)/20*sigma*5
    ef = ufss.efield_shapes.gaussian(t,sigma)

    uf2.set_efields([t,t,t,t],[ef,ef,ef,ef],[1,1,1,1],[(0,1),(1,0),(1,0)])
    uf2.set_polarization_sequence(['x','x','x','x'])
    uf2.set_t(0.1,dt=3)

    tau = uf2.t.copy() #dtau is the same as the dt for local oscillator
    T = np.array([0,100])
    uf2.set_pulse_delays([tau,T])

    sig = uf2.calculate_signal_all_delays()

    return sig

def calculate_2Q_spectrum_with_UF2(interaction_picture_shift):
    folder = os.path.join('fixtures','copolymer_P1')

    uf2 = ufss.DensityMatrices(os.path.join(folder,'open'),
                               conserve_memory=True,
                               detection_type='complex_polarization')

    uf2.interaction_picture_shift = interaction_picture_shift

    # defining the optical pulses in the RWA
    sigma = 1
    t = np.arange(-20,21)/20*sigma*5
    ef = ufss.efield_shapes.gaussian(t,sigma)

    uf2.set_efields([t,t,t,t],[ef,ef,ef,ef],[1,1,1,1],[(0,2),(2,0),(1,0)])
    uf2.set_polarization_sequence(['x','x','x','x'])
    uf2.set_t(0.1,dt=3)

    tau = uf2.t.copy() #dtau is the same as the dt for local oscillator
    T = np.array([0,100])
    uf2.set_pulse_delays([tau,T])

    sig = uf2.calculate_signal_all_delays()

    return sig

def calculate_integrated_spectrum_with_UF2(interaction_picture_shift):
    folder = os.path.join('fixtures','copolymer_P1')

    uf2 = ufss.DensityMatrices(os.path.join(folder,'open'),
                               conserve_memory=True,
                               detection_type='integrated_polarization')

    uf2.interaction_picture_shift = interaction_picture_shift

    # defining the optical pulses in the RWA
    sigma = 1
    t = np.arange(-20,21)/20*sigma*5
    ef = ufss.efield_shapes.gaussian(t,sigma)

    uf2.set_efields([t,t,t,t],[ef,ef,ef,ef],[1,1,1,1],[(0,1),(1,0),(1,0)])
    uf2.set_polarization_sequence(['x','x','x','x'])
    uf2.set_t(0.1,dt=3)

    tau = uf2.t.copy() #dtau is the same as the dt for local oscillator
    T = np.array([0,100])
    uf2.set_pulse_delays([tau,T])

    sig = uf2.calculate_signal_all_delays()

    return sig

def L2_norm(a,b):
    """Computes L2 norm of two nd-arrays, taking the b as the reference"""
    return np.sqrt(np.sum(np.abs(a-b)**2)/np.sum(np.abs(b)**2))
    

class test_shift(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        folder = os.path.join('fixtures','copolymer_P1')
        ufss.HLG.run(folder,conserve_memory=False)

    def test_1(self):
        sig1 = calculate_spectrum_with_UF2(True)
        sig2 = calculate_spectrum_with_UF2(False)
        # print(np.max(np.abs(sig1-sig2)))
        self.assertTrue(np.allclose(sig1,sig2))

    def test_2(self):
        sig1 = calculate_integrated_spectrum_with_UF2(True)
        sig2 = calculate_integrated_spectrum_with_UF2(False)
        # print(np.max(np.abs(sig1-sig2)))
        self.assertTrue(np.allclose(sig1,sig2))

    def test_3(self):
        sig1 = calculate_2Q_spectrum_with_UF2(True)
        sig2 = calculate_2Q_spectrum_with_UF2(False)
        # print(np.max(np.abs(sig1-sig2)))
        self.assertTrue(np.allclose(sig1,sig2))
    
if __name__ == '__main__':
    unittest.main()



#Standard python libraries
import os
import warnings
import time

#Dependencies - numpy, scipy,
import numpy as np
from scipy.sparse import load_npz, csr_matrix
from scipy.sparse.linalg import eigs

from ufss.perturbative_calculations import DPBaseClass
from ufss.perturbative_calculations import OpenBaseClass
from ufss.perturbative_calculations import RK_perturbative_container
from scipy.integrate import RK45

class RKOpenEngine(OpenBaseClass,DPBaseClass):
    """This class is designed to calculate perturbative wavepackets in the
        light-matter interaction given the eigenvalues of the unperturbed 
        hamiltonian and the material dipole operator evaluated in the
        eigenbasis of the unperturbed hamiltonian.

    Args:
        file_path (string): path to folder containing eigenvalues and the
            dipole operator for the system Hamiltonian
        detection_type (string): options are 'polarization' (default) or 'fluorescence'

"""
    def __init__(self,file_path,*,detection_type = 'polarization',
                 conserve_memory=False,method='Euler'):
        DPBaseClass.__init__(self,file_path,detection_type=detection_type)
        OpenBaseClass.__init__(self,detection_type = detection_type)
        self.slicing_time = 0
        self.interpolation_time = 0
        self.expectation_time = 0
        self.RK45_step_time = 0
        self.dipole_dot_rho_time = 0
        self.dipole_time = 0
        self.automation_time = 0
        self.diagram_to_signal_time = 0
        self.diagram_generation_counter = 0
        self.convolution_time = 0

        self.interaction_picture_calculations = True

        self.sparsity_threshold = 0.1

        self.conserve_memory = conserve_memory

        self.method = method

        self.load_L()

        self.set_rho_shapes()

        self.load_H_eigensystem()

        if not self.conserve_memory:
            self.load_mu()

        try:
            self.load_H_mu()
            self.H_mu_flag = True
        except FileNotFoundError:
            self.load_mu()
            self.H_mu_flag = False
            warnings.warn('Could not find H_mu, defaulting to L_mu for calculating expectation values, which is slower')

        self.atol = 1E-6
        self.rtol = 1E-5

        self.set_maximum_manifold()

    def set_efields(self,times_list,efields_list,centers_list,
                    phase_discrimination,*,reset_calculations = True,
                    plot_fields = False):
        
        DPBaseClass.set_efields(self,times_list,efields_list,centers_list,
                                 phase_discrimination,
                                 reset_calculations = reset_calculations,
                                 plot_fields = plot_fields)
        
        # Initialize unperturbed density matrix
        self.set_rho0_auto()

    def save_timing(self):
        save_dict = {'RKE_calculation_time':self.calculation_time}
        np.savez(os.path.join(self.base_path,'RKE_calculation_time.npz'),**save_dict)

    def add_impulsive_rhos(self,ra,rb):
        """Add two density matrix objects together
"""
        manifold_key = ra.manifold_key
        pdc = ra.pdc
        sim_a = ra.simultaneous
        sim_b = rb.simultaneous
        if sim_a != sim_b:
            raise Exception('Cannot add impulsive rhos if they do not have the same number of simultaneous interactions')
        else:
            simultaneous = sim_a
        pna = ra.pulse_number
        pnb = ra.pulse_number
        ta = self.pulse_times[pna]
        tb = self.pulse_times[pnb]
        if np.allclose(ta,tb):
            t = ra.t[1:2]
            rho_a = ra.f[:,1:2]
            rho_b = rb.f[:,1:2]
            pulse_number = max(pna,pnb)
        else:
            raise Exception('Cannot add two impulsive rhos unless they have the same value of t0')

        rho = rho_a + rho_b
        rho *= 2 * simultaneous
        
        rab=RK_perturbative_container(t,rho,pulse_number,manifold_key,pdc,
                                      simultaneous = simultaneous)
        rab.one_time_step_function = self.one_time_step_function
        return rab
        

    def add_rhos(self,ra,rb):
        """Add two density matrix objects together
"""
        if not np.allclose(ra.pdc,rb.pdc):
            raise Exception('Cannot add density matrices with different phase-discrimination conditions')
        if 'all_manifolds' in self.manifolds:
            pass
        else:
            if ra.manifold_key != rb.manifold_key:
                raise Exception('Cannot add perturbative_container objects that exist in different manifolds')
            
        if ra.impulsive:
            return self.add_impulsive_rhos(ra,rb)
        
        manifold_key = ra.manifold_key
        
        pdc = ra.pdc
        
        tmin = min(ra.t[0],rb.t[0])
        tmax = max(ra.t[-1],rb.t[-1])
        dt = min(ra.t[1]-ra.t[0],rb.t[1]-rb.t[0])
        t = np.arange(tmin,tmax+dt*0.9,dt)
        rho = ra(t) + rb(t)

        pna = ra.pulse_number
        pnb = rb.pulse_number
        pulse_number = max(pna,pnb)
        
        rab = RK_perturbative_container(t,rho,pulse_number,manifold_key,pdc)
        rab.one_time_step_function = self.one_time_step_function
        return rab

    def load_L(self):
        """Load in known eigenvalues. Must be stored as a numpy archive file,
with keys: GSM, SEM, and optionally DEM.  The eigenvalues for each manifold
must be 1d arrays, and are assumed to be ordered by increasing energy. The
energy difference between the lowest energy ground state and the lowest 
energy singly-excited state should be set to 0
"""
        L_save_name = os.path.join(self.base_path,'L.npz')
        try:
            with np.load(L_save_name,allow_pickle=True) as L_archive:
                self.L = dict()
                for key in L_archive.keys():
                    L = L_archive[key]
                    if L.dtype == np.dtype('O'):
                        self.L[key] = L[()]
                    else:
                        if self.check_sparsity(L):
                            self.L[key] = csr_matrix(L)
                        else:
                            self.L[key] = L
        except:
            self.L = {'all_manifolds':load_npz(L_save_name)}
        self.manifolds = list(self.L.keys())

        if len(self.manifolds[0]) == 2:
            self.convert_manifold_keys()

    def convert_manifold_keys(self):
        old_manifolds = self.manifolds
        new_manifolds = []
        for k,b in self.manifolds:
            new_key = ''.join([k,',',b])
            new_manifolds.append(new_key)
        for old_key,new_key in zip(old_manifolds,new_manifolds):
            self.L[new_key] = self.L.pop(old_key)

        self.manifolds = new_manifolds
        
    def dL(self,t,rho):
        try:
            L = self.L['all_manifolds']
        except KeyError:
            L = self.L[rho.manifold_key]
        return L.dot(rho)

    def get_dL_manual(self,manifold_key):
        try:
            L = self.L['all_manifolds']
        except KeyError:
            L = self.L[manifold_key]

        def L_fun(t,rho):
            return L.dot(rho)

        return L_fun

    def one_time_step_function(self,rho0,t0,tf,*,manifold_key = None):
        num_steps = 0
        if manifold_key == None:
            rk45 = RK45(self.dL,t0,rho0,tf,atol=self.atol,rtol=self.rtol)
        else:
            dL = self.get_dL_manual(manifold_key)
            rk45 = RK45(dL,t0,rho0,tf,atol=self.atol,rtol=self.rtol)
        if rk45.direction == 1:
            while rk45.t < tf:
                rk45.step()
                num_steps += 1
        else:
            while rk45.t > tf:
                rk45.step()
                num_steps += 1
        rho_final = rk45.y
        return rho_final

    def get_bottom_eigenvector(self):
        try:
            L = self.L['all_manifolds']
        except KeyError:
            L = self.L['0,0']
        if L.shape == (1,1):
            e = L[0,0]
            ev = np.array([[1]])
        else:
            e, ev = eigs(L,k=1,which='SM',maxiter=10000)
        if e.size == 1 and np.allclose(e,0):
            pass
        else:
            raise Exception('Smallest magnitude eigenvalue of L is {}. L must have a single stationary state for this code to work'.format(e))
        v = ev[:,0]
        H_size = int(np.sqrt(v.size))
        rho = v.reshape((H_size,H_size))
        trace = rho.trace()
        v = v/trace # Need to start with a trace 1 object
        return v

    def set_rho0_auto(self):
        try:
            rho0 = np.load(os.path.join(self.base_path,'rho0.npy'))
        except FileNotFoundError:
            rho0 = self.get_bottom_eigenvector()
        t = np.array([-np.inf,0,np.inf])
        rho0 = rho0[:,np.newaxis] * np.ones((rho0.size,t.size))
        pulse_number = None
        manifold_key = '0,0'
        pdc = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)
        
        self.rho0 = RK_perturbative_container(t,rho0,pulse_number,
                                              manifold_key,pdc,
                                              interp_kind = 'zero')

    def set_rho0_manual(self,rho0):
        t = np.array([-np.inf,0,np.inf])
        rho0 = rho0[:,np.newaxis] * np.ones((rho0.size,t.size))
        pulse_number = None
        manifold_key = '0,0'
        pdc = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)
        
        self.rho0 = RK_perturbative_container(t,rho0,pulse_number,
                                              manifold_key,pdc,
                                              interp_kind = 'zero')

    def set_rho_shapes(self):
        self.rho_shapes = dict()
        if 'all_manifolds' in self.manifolds:
            L_size = self.L['all_manifolds'].shape[0]
            H_size = int(np.sqrt(L_size))
            self.rho_shapes['all_manifolds'] = (H_size,H_size)
        else:
            H_sizes = dict()
            for key in self.manifolds:
                ket_key, bra_key = key.split(',')
                if ket_key == bra_key:
                    L_size = self.L[key].shape[0]
                    H_size = int(np.sqrt(L_size))
                    H_sizes[ket_key] = H_size
            for key in self.manifolds:
                ket_key, bra_key = key.split(',')
                ket_size = H_sizes[ket_key]
                bra_size = H_sizes[bra_key]
                self.rho_shapes[key] = (ket_size,bra_size)

    def get_rho(self,t,rho_obj,*,reshape=True,original_L_basis=True):
        manifold_key = rho_obj.manifold_key
        
        try:
            rho_shape = self.rho_shapes['all_manifolds']
        except KeyError:
            rho_shape = self.rho_shapes[manifold_key]
        rho_shape = rho_shape + (t.size,)
        
        rho = rho_obj(t)
        if reshape:
            rho = rho.reshape(rho_shape)
            
        return rho

    def load_mu(self):
        """Load the precalculated dipole overlaps.  The dipole operator must
            be stored as a .npz file, and must contain at least one array, each with three 
            indices: (new manifold index, old manifold eigenfunction, 
            cartesian coordinate)."""
        try:
            file_name = os.path.join(self.base_path,'mu_site_basis.npz')
            with np.load(file_name) as mu_archive:
                self.mu = {key:mu_archive[key] for key in mu_archive.keys()}
        except FileNotFoundError:
            try:
                file_name = os.path.join(self.base_path,'mu_original_L_basis.npz')
                with np.load(file_name) as mu_archive:
                    self.mu = {key:mu_archive[key] for key in mu_archive.keys()}
            except FileNotFoundError:
                file_name = os.path.join(self.base_path,'mu.npz')
                with np.load(file_name) as mu_archive:
                    self.mu = {key:mu_archive[key] for key in mu_archive.keys()}
        sparse_flags = []
        for key in self.mu.keys():
            mu_2D = np.sum(np.abs(self.mu[key])**2,axis=-1)
            sparse_flags.append(self.check_sparsity(mu_2D))
        sparse_flags = np.array(sparse_flags)
        if np.allclose(sparse_flags,True):
            self.sparse_mu_flag = True
        else:
            self.sparse_mu_flag = False

        for key in self.mu.keys():
            mu_x = self.mu[key][...,0]
            mu_y = self.mu[key][...,1]
            mu_z = self.mu[key][...,2]

            if self.sparse_mu_flag:
                self.mu[key] = [csr_matrix(mu_x),csr_matrix(mu_y),csr_matrix(mu_z)]
            else:
                self.mu[key] = [mu_x,mu_y,mu_z]

        if 'all_manifolds' in self.manifolds:
            pass
        elif ',' in key:
            pass
        else:
            self.convert_mu_keys()
        
    ### Tools for recursively calculating perturbed density maatrices using TDPT

    def rho_matrix_to_L_vector(self,rho,manifold_key,*,
                                diagonal_L_basis=False):
        t_size = rho.shape[-1]
        rho_vec_size = np.prod(rho.shape[:-1])
        rho = rho.reshape(rho_vec_size,t_size)
        
        return rho

    def rho_L_vector_to_matrix(self,rho,manifold_key):
        try:
            rho_shape = self.rho_shapes['all_manifolds']
        except:
            rho_shape = self.rho_shapes[manifold_key]
        tsize = rho.shape[-1]
        rho_shape = rho_shape + (tsize,)
        rho = rho.reshape(rho_shape)

        return rho

    def mu_dot_rho(self,t,rho_in,pulse_number,ket_flag=True,up_flag=True,
                   next_manifold_mask = None,mu_type='L'):
        old_manifold_key = rho_in.manifold_key
        new_manifold_key = self.get_new_manifold_key(old_manifold_key,
                                                     ket_flag,up_flag)

        mu = self.get_mu(pulse_number,old_manifold_key,
                         new_manifold_key,ket_flag=ket_flag,
                         up_flag=up_flag,mu_type=mu_type)
            
        if mu_type=='H':
            rho =self.get_rho(t,rho_in,reshape=True)

            rho = self.H_mu_dot_rho(mu,rho,ket_flag=ket_flag)
            
            rho = self.rho_matrix_to_L_vector(rho,new_manifold_key)
            
        elif mu_type=='L':
            rho =self.get_rho(t,rho_in,reshape=False)
            rho = mu.dot(rho)
        else:
            raise Exception("mu_type must be 'H' or 'L'")
            
        return rho

    def H_mu_dot_rho(self,mu,rho,ket_flag=True):
        if ket_flag:
            ein_logic = 'ij,jkl'
            operators = (mu,rho)
        else:
            ein_logic = 'ijl,jk'
            operators = (rho,mu)
            
        if 'all_manifolds' in self.manifolds:
            if len(self.rho_shapes['all_manifolds']) == 3:
                if ket_flag:
                    ein_logic = 'ij,ajkl'
                else:
                    ein_logic = 'aijl,jk'

        return np.einsum(ein_logic,*operators)

    def mu_exp_val(self,t,rho_in,pulse_number,ket_flag=True,up_flag=False,
                   mu_type='H'):
        old_manifold_key = rho_in.manifold_key
        new_manifold_key = self.get_new_manifold_key(old_manifold_key,
                                                     ket_flag,up_flag)

        if not self.H_mu_flag:
            if mu_type == 'H':
                warnings.warn('Using L_mu for expectation values because H_mu is not available')
                mu_type = 'L'
        
        mu = self.get_mu(pulse_number,old_manifold_key,
                                     new_manifold_key,ket_flag=ket_flag,
                                     up_flag=up_flag,mu_type=mu_type)
            
        if mu_type=='H':
            rho =self.get_rho(t,rho_in,reshape=True)
            if 'all_manifolds' in self.manifolds:
                if len(self.rho_shapes['all_manifolds']) == 3:
                    rho = rho[0,...]

            if ket_flag:
                exp_val = np.einsum('ij,jil',mu,rho)
            else:
                exp_val = np.einsum('ijl,ji',rho,mu)
            
        else:
            rho =self.get_rho(t,rho_in,reshape=False)
            rho = mu.dot(rho)

            rho = self.rho_L_vector_to_matrix(rho,new_manifold_key)
            exp_val = np.einsum('iik',rho)
            
        return exp_val

    def next_order(self,rho_in,*,ket_flag=True,up_flag=True,
                   next_manifold_mask = None,pulse_number = 0):
        """This function connects rho_p to rho_pj^(*) using a DFT convolution algorithm.

        Args:
            rho_in (perturbative_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold
        
        Return:
            rho_dict (perturbative_container): next-order density matrix
"""
        ### Generic
        tnext_order0 = time.time()
        if ket_flag == up_flag:
            # Rotating term excites the ket and de-excites the bra
            conjugate_flag = False
        else:
            # Counter-rotating term
            conjugate_flag = True

        if rho_in.pulse_number == None:
            old_pulse_time = -np.inf
        else:
            old_pulse_time = self.pulse_times[rho_in.pulse_number]
        pulse_time = self.pulse_times[pulse_number]

        if np.isclose(old_pulse_time,pulse_time):
            simultaneous = rho_in.simultaneous + 1
        else:
            simultaneous = 1

        t = self.efield_times[pulse_number] + pulse_time
        dt = self.dts[pulse_number]
        old_manifold_key = rho_in.manifold_key
        new_manifold_key = self.get_new_manifold_key(old_manifold_key,
                                                     ket_flag,up_flag)

        input_pdc = rho_in.pdc
        output_pdc = input_pdc.copy()
        if conjugate_flag:
            output_pdc[pulse_number][1] += 1
        else:
            output_pdc[pulse_number][0] += 1

        center = self.centers[pulse_number]
        efield = self.efields[pulse_number] * np.exp(-1j*center*t)

        if conjugate_flag:
            efield = np.conjugate(efield)

        if ket_flag:
            efield = 1j * efield
        else:
            efield = -1j * efield

        ##probably remove this, mu_type = 'H' always for DP methods, I think
        mu_type = 'H' if self.conserve_memory else 'L'

        rho = self.mu_dot_rho(t,rho_in,pulse_number,mu_type=mu_type,
                              ket_flag=ket_flag,up_flag=up_flag,
                              next_manifold_mask=next_manifold_mask)
        
        ### Generic
        rho = rho * efield[np.newaxis,:]

        ### DP methods

        M = t.size

        if M == 1:
            next_rho = rho
        else:
            next_rho = np.zeros(rho.shape,dtype='complex')

            dt = self.dts[pulse_number]
            
            if self.method == 'Euler':
                next_rho[:,0] = rho[:,0] * dt
                for i in range(1,t.size):
                    rho0 = next_rho[:,i-1]
                    t0 = t[i-1]
                    ta = time.time()
                    next_rho[:,i] = self.one_time_step_function(rho0,t0,t[i],manifold_key=new_manifold_key)

                    tb = time.time()
                    self.RK45_step_time += tb - ta

                    next_rho[:,i] += rho[:,i] * dt

            elif self.method == 'UF2':
                ref_time = pulse_time
                for i in range(t.size):
                    next_rho[:,i]=self.one_time_step_function(rho[:,i],t[i],
                                    ref_time,manifold_key=new_manifold_key)

                t0 = time.time()

                M = self.efield_times[pulse_number].size

                fft_convolve_fun = self.heaviside_convolve_list[pulse_number].fft_convolve2

                next_rho = fft_convolve_fun(next_rho,d=dt)

                t1 = time.time()
                self.convolution_time += t1-t0

                for i in range(t.size):
                    next_rho[:,i]=self.one_time_step_function(next_rho[:,i],
                                ref_time,t[i],manifold_key=new_manifold_key)

        rho_out = RK_perturbative_container(t,next_rho,pulse_number,
                                            new_manifold_key,output_pdc,
                                            simultaneous=simultaneous)
        rho_out.one_time_step_function = self.one_time_step_function
    
        return rho_out



import numpy as np

# Different electric field shapes that all approach a delta function in time

def gaussian(t,sigma):
    """t is time.  Gaussian pulse, with time-domain standard deviation sigma, 
    normalized to behave like a delta function as sigma -> 0"""
    pre = 1/(np.sqrt(2*np.pi)*sigma)
    return pre * np.exp(-t**2/(2*sigma**2))

def exponential(t,sigma):
    """t is time.  Exponential pulse, with frequency-domain decay constant 
    gamma, normalized to behave like a delta function as sigma -> 0"""
    pre = 1/(2*sigma)
    return pre * np.exp(-np.abs(t)/sigma)

def lorentzian(t,sigma):
    """t is time.  Lorentzian pulse, with time-domain width sigma, 
    normalized to behave like a delta function as sigma -> 0"""
    pre = sigma/2/np.pi
    return pre / (t**2 + (sigma/2)**2)

def sech(t,sigma):
    pre = 1/(np.pi*sigma)
    return pre * 1/np.cosh(t/sigma)

def constant_bandwidth_chirped_gaussian(t,bandwidth,chirp):
    sigma = np.sqrt(1+chirp**2)/bandwidth
    newsig = sigma/np.sqrt(1+1j*chirp)
    return gaussian(t,newsig)

def razorblade_gaussian(t,width,sigma):
    dt = t[1] - t[0]
    a = np.sinc(t*width/np.pi/2)*width/(2*np.pi)
    if sigma == 0:
        return a
    b = gaussian(t,sigma)
    return fftconvolve(a,b,mode='same')*dt



import os
import itertools
import numpy as np
from ufss.HLG import DiagonalizeHamiltonian, LiouvillianConstructor, DiagonalizeLiouvillian
import yaml

class manualH:
    def __init__(self,H,mu_ket_up,*,savedir='',RWA = True,for_RKE=False,
                 flourescence_yield=np.array([])):
        """Takes as input a Hamiltonian (H), and dipole operator (mu_ket_up) 
            and a dissipation super-operator D
"""
        self.H = self.convert_H_to_dict(H)
        self.mu = self.convert_mu_to_dict(mu_ket_up)
        if flourescence_yield.size == 0:
            fy_flag = False
        else:
            fy_flag = True
            self.fy = self.convert_fy_to_dict(flourescence_yield)
        self.check_mu_shape()
        if RWA:
            self.check_mu()

        self.for_RKE = for_RKE
        self.base_path = savedir
        self.save_path = os.path.join(self.base_path,'closed')
        os.makedirs(self.save_path,exist_ok=True)

        self.save_H()
        self.save_mu()
        self.make_params_file()
        if fy_flag:
            self.save_fy()

        self.run_closed()

    def make_params_file(self):
        filename = os.path.join(self.base_path,'params.yaml')
        params = {'ManualH':True}
        with open(filename,'w+') as yamlstream:
            yamlstream.write('### This file was generated by UFSS, do not modify unless you know what you are doing \n')
            yaml.dump(params,yamlstream)

    def convert_fy_to_dict(self,fy):
        if isinstance(fy,np.ndarray):
            fy_dict = {'all_manifolds':fy}
        elif isinstance(H,dict):
            fy_dict = fy
        else:
            raise Exception('Input H is not in a recognized format')

        return fy_dict

    def convert_H_to_dict(self,H):
        if isinstance(H,np.ndarray):
            Hdict = {'all_manifolds':H}
        elif isinstance(H,dict):
            Hdict = H
        else:
            raise Exception('Input H is not in a recognized format')

        return Hdict

    def convert_mu_to_dict(self,mu):
        if isinstance(mu,np.ndarray):
            mu_dict = {'up':mu}
        elif isinstance(mu,dict):
            mu_dict = mu
        else:
            raise Exception('Input mu is not in a recognized format')

        return mu_dict

    def check_mu_shape(self):
        for key in self.mu.keys():
            shape = self.mu[key].shape
            if len(shape) == 3:
                pass
            else:
                raise Exception("""Each component of mu must be a 3D ndarray,
                    where the third axis is the molecular frame cartesian 
                    axis""")

            if shape[-1] == 3:
                pass
            else:
                raise Exception("""Each component of mu must be a 3D ndarray,
                    where the third axis is the molecular frame cartesian 
                    axis, and must have 3 entries for x, y, z components""")

    def save_mu(self):
        np.savez(os.path.join(self.save_path,'mu_original_H_basis.npz'),
                 **self.mu)

    def save_H(self):
        np.savez(os.path.join(self.save_path,'H.npz'),**self.H)

    def save_fy(self):
        np.savez(os.path.join(self.save_path,'fluorescence_yield.npz'),**self.fy)
            
    def check_mu(self):
        if 'all_manifolds' not in self.H.keys():
            check = True
        else:
            for key in self.mu.keys():
                for i in range(3):
                    mu_i = self.mu[key][...,i]
                    check = np.allclose(mu_i,np.tril(mu_i))
                    if check:
                        pass
                    else:
                        break
        if not check:
            raise Exception("""dipole operator must be lower triangular 
                in order to ensure that the rotating wave approximation 
                is followed. If you know what you are doing, you can turn
                off this error by calling manual_H_and_D with RWA = 
                False""")

    def run_closed(self):
        if self.for_RKE:
            pass
        else:
            dh = DiagonalizeHamiltonian(self.base_path)

class manualD:
    def __init__(self,D,*,savedir='',H_eigenbasis=True,for_RKE=False,
                 conserve_memory=True):
        """Takes as input a Hamiltonian (H), and dipole operator (mu_ket_up) 
            and a dissipation super-operator D
"""
        self.base_path = savedir
        self.load_path = os.path.join(self.base_path,'closed')
        self.save_path = os.path.join(self.base_path,'open')
        os.makedirs(self.save_path,exist_ok=True)

        self.load_H(H_eigenbasis)

        self.D = self.convert_D_to_dict(D)
        self.D_operators_to_superoperator()
        self.flatten_all_Ds()

        self.set_L()
        self.save_L()
        self.modify_params_file(H_eigenbasis)
        if not conserve_memory:
            self.load_mu(H_eigenbasis)

            self.set_mu_Liouville_space()
            self.save_mu()
        if not for_RKE:
            dl = DiagonalizeLiouvillian(self.base_path,
                                        conserve_memory=conserve_memory)
            
    def modify_params_file(self,H_eigenbasis):
        filename = os.path.join(self.base_path,'params.yaml')
        params = {'ManualD':True,
                  'H_eigenbasis':H_eigenbasis}
        with open(filename,'a') as yamlstream:
            yaml.dump(params,yamlstream)

    def convert_D_to_dict(self,D):
        if isinstance(D,np.ndarray):
            Ddict = {'all_manifolds':D}
        elif isinstance(D,list):
            Ddict = {'all_manifolds':D}
        elif isinstance(D,dict):
            Ddict = D
        else:
            raise Exception('Input H is not in a recognized format')

        Dkeys = Ddict.keys()

        if 'all_manifolds' in Dkeys and len(Dkeys)==1:
            pass
        elif '00' in Dkeys:
            pass
        else:
            raise Exception('D dictionary keys are not recognized')

        return Ddict

    def D_operators_to_superoperator(self):
        for key in self.D.keys():
            if isinstance(self.D[key],np.ndarray):
                pass
            elif isinstance(self.D[key],list):
                LC = LiouvillianConstructor
                self.D[key] = LC.make_Liouvillian(self.D[key])
            else:
                raise Exception('D is not in a supported format')

    def flatten_all_Ds(self):
        for key in self.D.keys():
            if len(self.D[key].shape) == 4:
                self.D[key] = self.flatten_D(self.D[key])
            elif len(self.D[key].shape) == 2:
                pass
            else:
                raise Exception('D must be either a rank 2 or rank 4 tensor')
                

    def flatten_D(self,D):
        new_shape = (D.shape[0] * D.shape[1], D.shape[2] * D.shape[3])
        new_D = np.zeros(new_shape)

        for k,l in itertools.product(range(D.shape[2]),range(D.shape[3])):
            n = k*D.shape[3] + l
            new_rho = D[:,:,k,l]
            new_D[:,n] += new_rho.flatten()

        return new_D

    def load_H(self,H_eigenbasis):
        self.H = {}
        if H_eigenbasis:
            eigvals_path = os.path.join(self.load_path,'eigenvalues.npz')
            with np.load(eigvals_path) as arch:
                for key in arch.keys():
                    self.H[key] = np.diag(arch[key])
        else:
            H_path = os.path.join(self.load_path,'H.npz')
            with np.load(H_path) as arch:
                for key in arch.keys():
                    self.H[key] = arch[key]

        self.manifolds = list(self.H.keys())
                    
    def load_mu(self,H_eigenbasis):
        self.mu = {}
        if H_eigenbasis:
            mu_path = os.path.join(self.load_path,'mu.npz')
        else:
            mu_path = os.path.join(self.load_path,'mu_original_H_basis.npz')

        with np.load(mu_path) as arch:
            for key in arch.keys():
                self.mu[key] = arch[key]

    def make_U(self,ket_manifold,bra_manifold):
        H_ket = self.H[ket_manifold]
        H_bra = self.H[bra_manifold]
        LC = LiouvillianConstructor
        U_ins = LC.make_commutator_instructions2(-1j*H_ket,-1j*H_bra)
        U = LC.make_Liouvillian(U_ins,sparse=True)
        return U

    def set_L(self):
        self.L = {}
        if self.manifolds[0] == 'all_manifolds':
            L_key = 'all_manifolds'
            U = self.make_U(L_key,L_key)
            self.L[L_key] = U + self.D[L_key]
        else:
            for ket_man,bra_man in itertools.product(self.manifolds,
                                                     self.manifolds):
                L_key = ket_man + bra_man
                U = self.make_U(ket_man,bra_man)
                self.L[L_key] = U + self.D[L_key]

    def save_L(self):
        np.savez(os.path.join(self.save_path,'L.npz'),**self.L)

    def mu_key_to_manifold_keys(self,key):
        if self.manifolds[0] == 'all_manifolds':
            starting_key = 'all_manifolds'
            ending_key = 'all_manifolds'
        else:
            starting_key, ending_key = key.split('_to_')
        return starting_key, ending_key


    def make_mu_by_manifold_ket(self,old_manifold,change):
        i,j = old_manifold
        i2 = i + change
        j2 = j
        
        if i2 >= 0 and i2 <= self.PolyVib.maximum_manifold:
            pass
        else:
            return None, None

        if i2 > i:
            mu_key = str(i) + '_to_' + str(i2)
            mu = self.mu[mu_key]
        else:
            mu_key = str(i2) + '_to_' + str(i)
            mu = self.mu[mu_key].transpose(1,0,2)

        j_size = self.H[str(j)].shape[0]
        i_size = self.H[str(i)].shape[0]
        i2_size = self.H[str(i2)].shape[0]
        
        bra_eye = np.eye(j_size)
        old_key = str(i) + str(j)
        new_key = str(i2) + str(j2)
        
        mu_shape = (i2_size*j_size,i_size*j_size,3)
        new_mu = np.zeros(mu_shape,dtype='complex')
        for i in range(3):
            mu_i = np.kron(mu[:,:,i],bra_eye)
            new_mu[:,:,i] = mu_i

        if np.allclose(np.imag(new_mu),0):
            new_mu = np.real(new_mu)
        mu_key = old_key + '_to_' + new_key
        return mu_key, new_mu

    def make_mu_by_manifold_bra(self,old_manifold,change):
        i,j = old_manifold
        i2 = i
        j2 = j + change
        
        if j2 >= 0 and j2 <= self.PolyVib.maximum_manifold:
            pass
        else:
            return None, None

        if j2 > j:
            mu_key = str(j) + '_to_' + str(j2)
            mu = self.mu[mu_key].transpose(1,0,2)
        else:
            mu_key = str(j2) + '_to_' + str(j)
            mu = self.mu[mu_key]

        i_size = self.H[str(i)].shape[0]
        j_size = self.H[str(j)].shape[0]
        j2_size = self.H[str(j2)].shape[0]
        
        ket_eye = np.eye(i_size)
        old_key = str(i) + str(j)
        new_key = str(i2) + str(j2)
        
        mu_shape = (i_size*j2_size,i_size*j_size,3)
        new_mu = np.zeros(mu_shape,dtype='complex')
        for i in range(3):
            mu_i = np.kron(ket_eye,mu[:,:,i].T)
            new_mu[:,:,i] = mu_i

        if np.allclose(np.imag(new_mu),0):
            new_mu = np.real(new_mu)
        mu_key = old_key + '_to_' + new_key
        return mu_key, new_mu

    def make_mu_unseparable_manifolds(self,change,ket_flag):
        mu = self.mu['up']
        if ket_flag:
            mu_key = 'ket'
        else:
            mu = mu.transpose(1,0,2)
            mu_key = 'bra'
        if change == 1:
            mu_key += '_up'
        elif change == -1:
            mu = mu.transpose(1,0,2)
            mu_key += '_down'
        else:
            raise Exception('change must be either +1 or -1')
        H_size = self.H['all_manifolds'].shape[0]
        
        H_eye = np.eye(H_size)
        
        mu_shape = (H_size**2,H_size**2,3)
        new_mu = np.zeros(mu_shape,dtype='complex')
        for i in range(3):
            if ket_flag:
                mu_i = np.kron(mu[:,:,i],H_eye.T)
            else:
                mu_i = np.kron(H_eye,mu[:,:,i].T)
                
            new_mu[:,:,i] = mu_i

        if np.allclose(np.imag(new_mu),0):
            new_mu = np.real(new_mu)
            
        return mu_key, new_mu

    def append_mu_by_manifold(self,old_manifold,change,ket_flag):
        if ket_flag:
            f = self.make_mu_by_manifold_ket
        else:
            f = self.make_mu_by_manifold_bra
        key, mu = f(old_manifold,change)
        if key == None:
            pass
        else:
            self.mu_L_basis[key] = mu

    def set_mu_Liouville_space_unseparable_manifolds(self):
        self.mu_L_basis = dict()
        changes = [-1,1]
        for change,ket_flag in itertools.product(changes,[True,False]):
            mu_key, mu = self.make_mu_unseparable_manifolds(change,ket_flag)
            self.mu_L_basis[mu_key] = mu

    def set_mu_Liouville_space_separable_manifolds(self):
        self.mu_L_basis = dict()
        for i_key in self.manifolds:
            for j_key in self.manifolds:
                manifold = (int(i_key),int(j_key))
                self.append_mu_by_manifold(manifold,1,True)
                self.append_mu_by_manifold(manifold,-1,True)
                self.append_mu_by_manifold(manifold,1,False)
                self.append_mu_by_manifold(manifold,-1,False)

    def set_mu_Liouville_space(self):
        if self.manifolds[0] == 'all_manifolds':
            self.set_mu_Liouville_space_unseparable_manifolds()
        else:
            self.set_mu_Liouville_space_separable_manifolds()
                
    def save_mu(self):
        np.savez(os.path.join(self.save_path,'mu_original_L_basis.npz'),**self.mu_L_basis)



import numpy as np
import ufss
from ufss.signals import SpectroscopyBase
import matplotlib.pyplot as plt
import warnings

class Spectroscopy2D(SpectroscopyBase):

    def set_pulse_delays(self,pulse_delays):
        super().set_pulse_delays(pulse_delays)
        self.tau = pulse_delays[0]
        self.T = pulse_delays[1]

    def tau_dft(self,signal,*,mirror_tau_signal=True,add_heaviside=False,zero_pad=False):
        sig = signal.copy()
        tau = self.tau
        sig_shape = list(sig.shape)

        if mirror_tau_signal:
            if zero_pad:
                warnings.warn('Skipping zero_pad, not compatibile with option mirror_tau_signal')
            if add_heaviside:
                warnings.warn('Skipping add_heaviside, not compatibile with option mirror_tau_signal')

            new_tau_size = tau.size*2-1
            sig_shape[0] = new_tau_size
            new_sig = np.zeros(sig_shape,dtype='complex')
            new_sig[new_tau_size//2:,...] = sig
            sig_flipped = sig[::-1,...]
            new_sig[:new_tau_size//2,...] = sig_flipped[:-1,...]
            new_tau = np.zeros(tau.size*2-1,dtype=tau.dtype)
            new_tau[new_tau_size//2:] = tau
            new_tau[:new_tau_size//2+1] = -tau[::-1]
            tau = new_tau
            sig = new_sig

        else:
            if add_heaviside:
                sig[0,...] *= 0.5
            if zero_pad:
                new_tau_size = tau.size*2-1
                sig_shape[0] = new_tau_size
                new_sig = np.zeros(sig_shape,dtype='complex')
                new_sig[new_tau_size//2:,...] = sig
                new_tau = np.zeros(tau.size*2-1,dtype=tau.dtype)
                new_tau[new_tau_size//2:] = tau
                new_tau[:new_tau_size//2+1] = -tau[::-1]
                tau = new_tau
                sig = new_sig
        
        if self.rephasing:
            wtau,signal_ft = ufss.signals.SignalProcessing.ft1D(tau,sig,
                                                                axis=0)
            signal_ft = signal_ft/(2*np.pi)
        else:
            wtau,signal_ft = ufss.signals.SignalProcessing.ift1D(tau,sig,
                                                                 axis=0)
        self.wtau = wtau
        return wtau, signal_ft

    # def tau_dft(self,signal,*,add_heaviside=True,zero_pad=True):
    #     sig = signal.copy()
    #     if add_heaviside:
    #         sig[0,...] *= 0.5
    #     tau = self.tau
    #     t = self.engine.t
    #     if zero_pad:
    #         sig_shape = list(sig.shape)
    #         new_tau_size = tau.size*2-1
    #         sig_shape[0] = new_tau_size
    #         new_sig = np.zeros(sig_shape,dtype='complex')
    #         new_sig[new_tau_size//2:,...] = sig
    #         new_tau = np.zeros(tau.size*2-1,dtype=tau.dtype)
    #         new_tau[new_tau_size//2:] = tau
    #         new_tau[:new_tau_size//2+1] = -tau[::-1]
    #         tau = new_tau
    #         sig = new_sig
    #     if self.rephasing:
    #         wtau,signal_ft = ufss.signals.SignalProcessing.ft1D(tau,sig,
    #                                                             axis=0)
    #         signal_ft = signal_ft/(2*np.pi)
    #     else:
    #         wtau,signal_ft = ufss.signals.SignalProcessing.ift1D(tau,sig,
    #                                                              axis=0)
    #     self.wtau = wtau
    #     return wtau, signal_ft

    def get_signal(self,lam,*,tau_dft = True):
        """lam is the perturbative parameter
        Args:
            lam (float) : unit-less electric field amplitude scaling factor
"""
        lam_list = [lam,lam,1]
        signal = super().get_signal(lam_list)

        if tau_dft == True:
            wtau, signal = self.tau_dft(signal)

        return signal

    def get_signal_order(self,order,*,tau_dft = True):
        signal = super().get_signal_order(order)

        if tau_dft == True:
            wtau, signal = self.tau_dft(signal)

        return signal

    def plot_signal(self,lam,T_index):
        signal = self.get_signal(lam)[:,T_index,:]
        wt = self.engine.w
        ufss.signals.plot2D(self.wtau,wt,signal,part='real')
        plt.xlabel(r'$\omega_{\tau}$')
        plt.ylabel(r'$\omega_t - \omega_c$')

    def plot_order(self,order,T_index):
        signal = self.get_signal_order(order)[:,T_index,:]
        wt = self.engine.w
        ufss.signals.plot2D(self.wtau,wt,signal,part='real')
        plt.xlabel(r'$\omega_{\tau}$')
        plt.ylabel(r'$\omega_t - \omega_c$')

class Rephasing(Spectroscopy2D):

    def __init__(self,file_name,*,engine_name = 'UF2',conserve_memory=False):
        super().__init__(file_name,engine_name = engine_name,
                         detection_type='complex_polarization',
                         conserve_memory=conserve_memory)

        pdc = ((0,1),(1,0),(1,0))
        self.set_phase_discrimination(pdc)
        self.save_name = 'rephasing_signals'
        self.rephasing = True #used for tau Fourier convention

    def set_highest_order(self,highest_order):
        pulse_orders = (highest_order-2,highest_order-2,1)
        self.engine.include_higher_orders(pulse_orders)

class NonRephasing(Spectroscopy2D):

    def __init__(self,file_name,*,engine_name = 'UF2',conserve_memory=False):
        super().__init__(file_name,engine_name = engine_name,
                         detection_type='complex_polarization',
                         conserve_memory=conserve_memory)

        pdc = ((1,0),(0,1),(1,0))
        self.set_phase_discrimination(pdc)
        self.save_name = 'nonrephasing_signals'
        self.rephasing = False #used for tau Fourier convention

    def set_highest_order(self,highest_order):
        pulse_orders = (highest_order-2,highest_order-2,1)
        self.engine.include_higher_orders(pulse_orders)

class PumpPumpProbe(Spectroscopy2D):
    def __init__(self,file_name,*,engine_name = 'UF2',conserve_memory=False,
                 include_linear=False,include_wtau_DC_terms=False):
        super().__init__(file_name,engine_name=engine_name,
                         conserve_memory=conserve_memory)
        self.pdc = ((0,0),(0,0),(1,0))
        self.save_name = 'pump_pump_probe_signals'
        self.rephasing = True #used for tau Fourier convention
        self.include_linear = include_linear
        self.include_wtau_DC_terms = include_wtau_DC_terms

    def set_identical_gaussians(self,sigma,center,*,M=51,Delta=10):
        t = np.linspace(-sigma,sigma,num=M,endpoint=True)*Delta/2
        f = ufss.gaussian(t,sigma)
        pump = f * np.exp(-1j*center*t)
        probe = f
        times = [t,t,t]
        efields = [pump,pump,probe]
        centers = [0,0,center]
        self.engine.set_efields(times,efields,centers,self.pdc)
        self.engine.set_polarization_sequence(['x']*3)

    def set_impulsive_pulses(self,center):
        self.engine.set_impulsive_pulses(self.pdc)
        self.engine.set_polarization_sequence(['x']*3)
        self.engine.centers = [0,0,center]
            
    def set_highest_order(self,highest_order):
        pulse_orders = [highest_order-1,highest_order-1,1]
        lowest_order_pdcs = []
        for i in range(highest_order//2):
            r,nr = self.pump_pump_probe_pdcs(i+1)
            lowest_order_pdcs += [r,nr]

        self.engine.include_higher_orders(pulse_orders)
        for pdc in lowest_order_pdcs:
            self.engine.include_additional_pdc(pdc)
        if self.include_wtau_DC_terms:
            if not self.include_linear:
                self.engine.signal_pdcs.remove(self.pdc)
        else:
            self.engine.remove_signal_pdcs(self.pdc)
            if self.include_linear:
                self.engine.signal_pdcs.append(self.pdc)

    def pump_pump_probe_pdcs(self,n):
        """Returns (non-)rephasing lowest order nQ pdc"""
        r = ((0,n),(n,0),(1,0))
        nr = ((n,0),(0,n),(1,0))
        return r,nr
    
    def load(self,file_name,*,use_base_path=True):
        self.engine.load(file_name,pulse_delay_names=[],
                         use_base_path=use_base_path)
        self.tau = self.engine.all_pulse_delays[0]
        self.T = self.engine.all_pulse_delays[1]
        




from ufss.diagram_automation import DiagramGenerator
from ufss.composite_diagrams import CompositeDiagrams
from ufss.signals import UF2Wavefunctions, UF2DensityMatrices, RKWavefunctions, RKDensityMatrices
import ufss.signals as signals
import ufss.HLG as HLG
from ufss.efield_shapes import *
from ufss.convergence_tester import efieldConvergence, Convergence

## Backwards compatibility
Wavepackets = UF2Wavefunctions
DensityMatrices = UF2DensityMatrices
RKE_DensityMatrices = RKDensityMatrices


try:
    import pyfftw
    from .heaviside_convolve_pyfftw import HeavisideConvolve
except ImportError:
    from .heaviside_convolve_numpy import HeavisideConvolve
from .general_base_class import UF2BaseClass, DPBaseClass
from .open_base_class import OpenBaseClass
from .closed_base_class import ClosedBaseClass
from .containers import perturbative_container, RK_perturbative_container
from .containers import ChebPoly, cheb_perturbative_container
from .UF2_open_core import UF2OpenEngine
from .UF2_core import UF2ClosedEngine
from .RK_open_core import RKOpenEngine
from .RK_core import RKClosedEngine




"""Adapted from:
A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    
setup(
    name='ufss',

    version='0.2.4',

    description='Package for simulating nonlinear optical spectra',

    long_description=long_description,

    long_description_content_type='text/markdown',

    url='https://github.com/peterarose/ufss/',

    author='Peter A. Rose and Jacob J. Krich',

    author_email='peter.rose56@gmail.com',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Science/Research',
        
        'Topic :: Scientific/Engineering :: Physics',
        
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='nonlinear optical spectroscopy',

    packages=find_packages(exclude=['docs']),

    python_requires='>=3.5',

    install_requires=['numpy','matplotlib','scipy>=1','pyx','pyyaml'],

)



import unittest
import numpy as np
import os

import ufss

def g(t,sigma):
    """t is time.  Gaussian pulse, with time-domain standard deviation sigma, 
        normalized to behave like a delta function as sigma -> 0"""
    pre = 1/(np.sqrt(2*np.pi)*sigma)
    return pre * np.exp(-t**2/(2*sigma**2))

def calculate_spectrum_with_UF2():
    folder = os.path.join('fixtures','v_3LS')

    re = ufss.DensityMatrices(os.path.join(folder,'open'),
                              detection_type='complex_polarization')

    # defining the optical pulses in the RWA
    M = 25 # number of points used to resolve optical pulses
    Delta = 6 # pulse interval
    t = np.linspace(-Delta/2,Delta/2,num=M)
    dt = t[1] - t[0]

    sigma = 1
    ef = g(t,sigma)

    # Smallwood et al. use a delta function for the local oscillator
    lo_dt = 0.25 #### This must never change, because it defines t and tau, by default
    lo_t =  np.arange(-5,5.2,lo_dt)    
    lo_dt = lo_t[1] - lo_t[0]
    lo = np.zeros(lo_t.size,dtype='float')
    lo[lo.size//2] = 1/lo_dt


    re.set_polarization_sequence(['x','x','x','x'])

    re.set_efields([t,t,t,lo_t],[ef,ef,ef,lo],[0,0,0,0],[(0,1),(1,0),(1,0)])

    gamma_dephasing = 0.2/sigma
    re.gamma_res = 20
    re.set_t(gamma_dephasing)
    re.pulse_times = [0,0,0]

    tau = re.t.copy() #dtau is the same as the dt for local oscillator
    T = np.arange(0,1,1)
    re.set_pulse_delays([tau,T])

    time_ordered_diagrams = [(('Bu', 0), ('Ku', 1), ('Ku', 2)), (('Bu', 0), ('Ku', 1), ('Bd', 2)), (('Bu', 0), ('Bd', 1), ('Ku', 2))]

    sig = re.calculate_diagrams_all_delays(time_ordered_diagrams)

    ift = ufss.signals.SignalProcessing.ift1D
    
    wtau, sig_ft = ift(tau,sig,zero_DC=False,axis=0)

    return sig_ft

def L2_norm(a,b):
    """Computes L2 norm of two nd-arrays, taking the b as the reference"""
    return np.sqrt(np.sum(np.abs(a-b)**2)/np.sum(np.abs(b)**2))
    

class test_against_smallwood(unittest.TestCase):
    def test(self):
        sig_ft = calculate_spectrum_with_UF2()
        analytical_sig_path = os.path.join('fixtures','v_3LS',
                                           'analytical_signal.npy')
        analytical_sig = np.load(analytical_sig_path)
        # signal definition has changed since this test was first made
        # signals in UFSS are now a factor of pi smaller, and complex
        # signals have a factor of -1 relative to the original implementation
        diff = L2_norm(-np.pi*sig_ft[:,0,:],analytical_sig[:,0,:])
        self.assertTrue(diff < 0.007)
    


if __name__ == '__main__':
    unittest.main()



from .Hamiltonians import LadderOperators, DisplacedAnharmonicOscillator, PolymerVibrations,DiagonalizeHamiltonian

from .general_Liouvillian_classes import LiouvillianConstructor

from .Redfield_Liouvillians import RedfieldConstructor,SecularRedfieldConstructor,DiagonalizeLiouvillian

from .DiabaticLindblad_Liouvillians import OpenPolymer, OpenPolymerVibrations

from .params_converter import convert

from .manual_H_and_D_input import manualH, manualD

from .manual_L_input import ManualL

from .run_HLG import run

from .make_simple_systems import make_2LS, make_3LS


#Standard python libraries
import os
import warnings
import copy
import time
import itertools
import functools

#Dependencies - numpy, scipy, matplotlib, pyfftw
import numpy as np
import pyx

class DiagramDrawer:
    """This class is used to draw double-sided Feynman diagrams and save them as pdf files
"""
    def __init__(self):
        self.draw_functions = {'Ku':self.draw_Ku,'Kd':self.draw_Kd,'Bu':self.draw_Bu,'Bd':self.draw_Bd}
        self.pulse_labels = ['a','b','c','d','e','f','g','h','i','j','k','l','m',
                             'n','o','p','q','r','s','t','u','v','w','x','y','z']

        self.include_state_labels = True
        self.include_emission_arrow = True
        self.include_pulse_labels = True
        self.sawtooth = None
        self.brightness = 0
        self.saturation = 1
        self.hues = []
        self.vertical_scale = 1
        self.diagram_size = 'tall'
        self.rails_hsb = (1,1,0)

    def set_pulse_colors(self,hues='auto',brightness='auto',saturation='auto'):
        """Set the color used to draw each pulse arrow and text. If called with
            no arguments, it will automatically asign colors to each pulse. If
            this method is never called, all pulses will be drawn in black.

        Args:
            hues (list) : list of hue values for each pulse
            brightness (list) : list of brightness values for each pulse
            saturation (list) : list of saturation values for each pulse
            """
        n = self.pdc.shape[0]
        if hues == 'auto':
            hues = np.arange(n)/n + 0.1
            self.hues = np.mod(hues,1)
        else:
            self.hues = hues
        if brightness == 'auto':
            self.brightness = [1]*n
        else:
            self.brightness = brightness
        if saturation == 'auto':
            self.saturation = [1]*n
        else:
            self.saturation = saturation
    
    def color_function(self,pulse_num):
        """Get the color of each pulse as a pyx color object (default is black)

        Args:
            pulse_num (int) : pulse number
        Returns:
            pyx.color.hsb : pyx color object for the given pulse
        """
        if self.hues == []:
            return pyx.color.hsb(1,self.saturation,self.brightness)
        else:
            return pyx.color.hsb(self.hues[pulse_num],self.saturation[pulse_num],
                                 self.brightness[pulse_num])

    def save_diagram(self,diagram,*,folder_name = ''):
        """Saves a pdf of an individual Feynman diagram

        Args:
            diagram (tuple) : Feynman diagram
            folder_name (str, optional) : relative or absolute path to folder 
                where diagram will be saved
        """
        os.makedirs(folder_name,exist_ok=True)
        filename = self.draw_diagram(diagram)
        self.c.writePDFfile(os.path.join(folder_name,filename))

    def save_diagrams(self,diagrams,*,folder_name=''):
        """Saves pdf files for each Feynman diagram supplied

        Args:
            diagrams (list) : list of Feynman diagrams
            folder_name (str,optional) : relative or absolute path to folder
                where diagram will be saved
        """
        for diagram in diagrams:
            self.save_diagram(diagram,folder_name=folder_name)

    def draw_diagram(self,diagram):
        """Draws Feynman diagram for viewing in a Jupyter notebook

        Args:
            diagram (tuple) : Feynman diagram
        Returns:
            str : unique filename for this diagram (used only for saving)
        """
        if self.diagram_size == 'medium':
            filename = self.draw_medium_diagram(diagram)
        elif self.diagram_size == 'compact':
            filename = self.draw_compact_diagram(diagram)
        elif self.diagram_size == 'tall':
            filename = self.draw_tall_diagram(diagram)
        else:
            raise Exception('Attribute diagram_size has an invalid string')

        return filename
        
    def draw_tall_diagram(self,diagram):
        """Draws Feynman diagram for viewing in a Jupyter notebook. Tall 
            indicates that the diagrams will be spread out vertically. This
            makes diagrams take up a lot of space but ensures that arrows
            never overlap.

        Args:
            diagram (tuple) : Feynman diagram
        Returns:
            str : unique filename for this diagram (used only for saving)
        """
        self.vertical_scale = 1
        self.c = pyx.canvas.canvas()
        interaction_counter = 0
        self.diagram_drawer_manifolds = np.array([0,0],dtype='int')
        self.y0 = 0

        filename = ''
        ymin1 = 0
        if self.include_state_labels:
            ymin1 = -0.4
            self.c.stroke(pyx.path.line(0,ymin1,0,0),[pyx.color.hsb(*self.rails_hsb)])
            self.c.stroke(pyx.path.line(1,ymin1,1,0),[pyx.color.hsb(*self.rails_hsb)])
            k,b = list(self.diagram_drawer_manifolds)
            rho_str = r'$|{}\rangle\langle{}|$'.format(k,b)
            self.c.text(0.1,-0.25,rho_str)
            
        if self.sawtooth == 'bottom':
            ymin2 = ymin1 - 0.3
            self.c.stroke(pyx.path.line(0,ymin2,0,ymin1),[pyx.color.hsb(*self.rails_hsb)])
            self.c.stroke(pyx.path.line(1,ymin2,1,ymin1),[pyx.color.hsb(*self.rails_hsb)])
            self.draw_sawtooth(ymin2)
        
        for KB,n in diagram:
            self.draw_functions[KB](n,interaction_counter)
            filename += KB + str(n)
            interaction_counter += 1
        if interaction_counter % 2:
            if self.include_emission_arrow:
                self.draw_emission()
        if self.sawtooth == 'top':
            y = self.y0
            y_extra = y + 0.45
            self.c.stroke(pyx.path.line(0, y, 0, y_extra),[pyx.color.hsb(*self.rails_hsb)])
            self.c.stroke(pyx.path.line(1, y, 1, y_extra),[pyx.color.hsb(*self.rails_hsb)])
            self.draw_sawtooth(y_extra)
        return filename

    def draw_medium_diagram(self,diagram):
        """Draws Feynman diagram for viewing in a Jupyter notebook. Medium 
            indicates that the diagrams will be less spread out vertically, as
            commpared with "tall". Arrows may overlap to some extent.

        Args:
            diagram (tuple) : Feynman diagram
        Returns:
            str : unique filename for this diagram (used only for saving)
        """
        self.vertical_scale = 0.6
        self.c = pyx.canvas.canvas()
        interaction_counter = 0
        self.diagram_drawer_manifolds = np.array([0,0],dtype='int')
        self.y0 = 0

        filename = ''
        ymin1 = 0
        if self.include_state_labels:
            k,b = list(self.diagram_drawer_manifolds)
            rho_str = r'$|{}\rangle\langle{}|$'.format(k,b)
            self.c.text(0.1,0.15,rho_str)
            
        if self.sawtooth == 'bottom':
            ymin2 = ymin1 - 0.3
            self.c.stroke(pyx.path.line(0, ymin2, 0, ymin1),[pyx.color.hsb(*self.rails_hsb)])
            self.c.stroke(pyx.path.line(1, ymin2, 1, ymin1),[pyx.color.hsb(*self.rails_hsb)])
            self.draw_sawtooth(ymin2)
        
        for KB,n in diagram:
            self.draw_functions[KB](n,interaction_counter)
            filename += KB + str(n)
            interaction_counter += 1
        if interaction_counter % 2:
            if self.include_emission_arrow:
                yf = self.y0 + 1.1
                self.c.stroke(pyx.path.line(0, self.y0, 0, yf),[pyx.color.hsb(*self.rails_hsb)])
                self.c.stroke(pyx.path.line(1, self.y0, 1, yf),[pyx.color.hsb(*self.rails_hsb)])
                self.draw_emission()
                self.y0 = yf
        if self.sawtooth == 'top':
            y = self.y0
            y_extra = y + 0.45
            self.c.stroke(pyx.path.line(0, y, 0, y_extra),[pyx.color.hsb(*self.rails_hsb)])
            self.c.stroke(pyx.path.line(1, y, 1, y_extra),[pyx.color.hsb(*self.rails_hsb)])
            self.draw_sawtooth(y_extra)
        return filename

    def draw_compact_diagram(self,diagram):
        """Draws Feynman diagram for viewing in a Jupyter notebook. Compact 
            indicates that the diagrams will be very compact vertically. 
            Arrows may overlap a great deal. Typically should only be used
            with attribute "include_pulse_labels" set to False.

        Args:
            diagram (tuple) : Feynman diagram
        Returns:
            str : unique filename for this diagram (used only for saving)
        """
        self.vertical_scale = 0.3
        self.c = pyx.canvas.canvas()
        interaction_counter = 0
        self.diagram_drawer_manifolds = np.array([0,0],dtype='int')
        self.y0 = 0

        filename = ''
        ymin1 = 0
        if self.include_state_labels:
            k,b = list(self.diagram_drawer_manifolds)
            rho_str = r'$|{}\rangle\langle{}|$'.format(k,b)
            self.c.text(0.1,0.15,rho_str)
            
        if self.sawtooth == 'bottom':
            ymin2 = ymin1 - 0.3
            self.c.stroke(pyx.path.line(0, ymin2, 0, ymin1),[pyx.color.hsb(*self.rails_hsb)])
            self.c.stroke(pyx.path.line(1, ymin2, 1, ymin1),[pyx.color.hsb(*self.rails_hsb)])
            self.draw_sawtooth(ymin2)

        old_labels_flag = self.include_state_labels
        self.include_state_labels = False
        for KB,n in diagram:
            self.draw_functions[KB](n,interaction_counter)
            filename += KB + str(n)
            interaction_counter += 1
        
        self.include_state_labels = old_labels_flag
        if self.include_state_labels:
            yf = self.y0 + 0.75
            self.c.stroke(pyx.path.line(0, self.y0, 0, yf),[pyx.color.hsb(*self.rails_hsb)])
            self.c.stroke(pyx.path.line(1, self.y0, 1, yf),[pyx.color.hsb(*self.rails_hsb)])
            k,b = list(self.diagram_drawer_manifolds)
            rho_str = r'$|{}\rangle\langle{}|$'.format(k,b)
            self.c.text(0.1,self.y0 + 0.45,rho_str)
            self.y0 = yf

            if self.sawtooth == 'top':
                y = self.y0
                y_extra = y + 0.05
                self.c.stroke(pyx.path.line(0, y, 0, y_extra),[pyx.color.hsb(*self.rails_hsb)])
                self.c.stroke(pyx.path.line(1, y, 1, y_extra),[pyx.color.hsb(*self.rails_hsb)])
                self.draw_sawtooth(y_extra)
        else:
            if self.sawtooth == 'top':
                y = self.y0
                y_extra = y + 0.75
                self.c.stroke(pyx.path.line(0, y, 0, y_extra),[pyx.color.hsb(*self.rails_hsb)])
                self.c.stroke(pyx.path.line(1, y, 1, y_extra),[pyx.color.hsb(*self.rails_hsb)])
                self.draw_sawtooth(y_extra)
            
        if interaction_counter % 2:
            if self.include_emission_arrow:
                yf = self.y0 + 0.75
                self.c.stroke(pyx.path.line(0, self.y0, 0, yf),[pyx.color.hsb(*self.rails_hsb)])
                self.c.stroke(pyx.path.line(1, self.y0, 1, yf),[pyx.color.hsb(*self.rails_hsb)])
                self.draw_emission()

        return filename
        
    def display_diagram(self,diagram,*,exclude="image/png"):
        """Displays a diagram in a Jupyter notebook environment or similar

        Args:
            diagram (tuple) : Feynman diagram to be drawn, must be a list or 
                tuple of tuples
            exclude (str,optional) : MIME types to exclude when attempting to 
                display diagram
        """
        filename = self.draw_diagram(diagram)
        display(self.c,exclude=exclude)

    def display_diagrams(self,diagrams,*,exclude="image/png"):
        """Displays diagrams in a Jupyter notebook environment or similar

        Args:
            diagrams (list) : list of Feynman diagram to be drawn, each of 
                which must be a list or tuple of tuples
            exclude (str,optional) : MIME types to exclude when attempting to 
                display diagram
        """
        for diagram in diagrams:
            self.display_diagram(diagram,exclude=exclude)

    def draw_emission(self):
        """Draws the emission arrow at the top of a diagram"""
        change = np.array([-1,0])
        self.diagram_drawer_manifolds += change
        yf = self.y0 + self.vertical_scale
        self.double_sided(yf)
        x,y = (0,self.y0+0.5)
        x2,y2 = (x-.5,y+.5)
        self.c.stroke(pyx.path.line(x,y,x2,y2),[pyx.style.linestyle.dashed,pyx.deco.earrow(size=0.3)])

    def draw_sawtooth(self,y0):
        """Draws a sawtooth line indicating that the diagram has been truncated
        """
        num_teeth = 3
        x_len = 1/num_teeth/2
        y_len = x_len * np.tan(np.pi/3)
        x0 = 0
        y1 = y0 + y_len
        for i in range(num_teeth):
            x1 = x0 + x_len
            self.c.stroke(pyx.path.line(x0,y0,x1,y1),[pyx.color.hsb(*self.rails_hsb)])
            self.c.stroke(pyx.path.line(x1,y1,x1+x_len,y0),[pyx.color.hsb(*self.rails_hsb)])
            x0 = x1 + x_len
        return None

    def double_sided(self,yf):
        """Draws the double-sided rails for the diagram

        Args:
            yf (float) : height of Feynman diagram
        """
        self.c.stroke(pyx.path.line(0, self.y0, 0, yf),[pyx.color.hsb(*self.rails_hsb)])
        self.c.stroke(pyx.path.line(1, self.y0, 1, yf),[pyx.color.hsb(*self.rails_hsb)])
        
        if self.include_state_labels:
            k,b = list(self.diagram_drawer_manifolds)
            rho_str = r'$|{}\rangle\langle{}|$'.format(k,b)
            self.c.text(0.1,self.y0 + 0.75, rho_str)

    def K_circle(self,interaction_num,pulse_num):
        """Draws a circle on the ket-side rail of the diagram

        Args:
            interaction_num (int) : interaction number
            pulse_num (int) : pulse number
        """
        self.c.fill(pyx.path.circle(0,interaction_num+0.5,0.1),[self.color_function(pulse_num)])

    def B_circle(self,interaction_num,pulse_num):
        """Draws a circle on the bra-side rail of the diagram

        Args:
            interaction_num (int) : interaction number
            pulse_num (int) : pulse number
        """
        self.c.fill(pyx.path.circle(1,interaction_num+0.5,0.1),[self.color_function(pulse_num)])

    def right_arrow(self,x,y,interaction_num,pulse_num):
        """Draws an arrow pointing up and to the right, depicting an 
            interaction with the rotating term

        Args:
            interaction_num (int) : interaction number
            pulse_num (int) : pulse number
        """
        xf, yf = (x+0.5,y+0.5)
        self.c.stroke(pyx.path.line(x,y,xf,yf),[self.color_function(pulse_num),pyx.deco.earrow(size=0.3)])
        return xf,yf

    def left_arrow(self,x,y,interaction_num,pulse_num):
        """Draws an arrow pointing up and to the left, depicting an 
            interaction with the counter-rotating term

        Args:
            interaction_num (int) : interaction number
            pulse_num (int) : pulse number
        """
        xf,yf = (x-.5,y+.5)
        self.c.stroke(pyx.path.line(x,y,xf,yf),[self.color_function(pulse_num),pyx.deco.earrow(size=0.3)])
        return xf,yf

    def pulse_text(self,x,y,pulse_num):
        """Rights the pulse label next to the associated interaction arrow

        Args:
            x (float) : x-position on the diagram where the text will go
            y (float) : y-position on the diagram where the text will go
            pulse_num (int) : pulse number
        """
        if self.include_pulse_labels:
            text = self.pulse_labels[pulse_num]
            self.c.text(x,y,text)

    def draw_Bd(self,pulse_num,interaction_num):
        """Draws the bra-down (Bd) interaction, which is the rotating term
            interacting with the bra-side

        Args:
            pulse_num (int) : pulse number
            interaction_num (int) : interaction number
        """
        change = np.array([0,-1])
        self.diagram_drawer_manifolds += change
        m = interaction_num
        n = pulse_num
        yf = self.y0 + self.vertical_scale
        self.double_sided(yf)
        self.B_circle(self.y0,n)
        x,y = (1,self.y0+0.5)
        x2,y2 = self.right_arrow(x,y,self.y0,n)
        self.pulse_text(x2-0.1,y2-0.5,n)
        self.y0 = yf

    def draw_Bu(self,pulse_num,interaction_num):
        """Draws the bra-up (Bu) interaction, which is the counter-rotating
            term interacting with the bra-side

        Args:
            pulse_num (int) : pulse number
            interaction_num (int) : interaction number
        """
        change = np.array([0,1])
        self.diagram_drawer_manifolds += change
        m = interaction_num
        n = pulse_num
        yf = self.y0 + self.vertical_scale
        self.double_sided(yf)
        self.B_circle(self.y0,n)
        x,y = (1.5,self.y0)
        self.left_arrow(x,y,self.y0,n)
        self.pulse_text(x-0.1,y+0.1,n)
        self.y0 = yf

    def draw_Kd(self,pulse_num,interaction_num):
        """Draws the ket-down (Kd) interaction, which is the counter-rotating
            term interacting with the ket-side

        Args:
            pulse_num (int) : pulse number
            interaction_num (int) : interaction number
        """
        change = np.array([-1,0])
        self.diagram_drawer_manifolds += change
        m = interaction_num
        n = pulse_num
        yf = self.y0 + self.vertical_scale
        self.double_sided(yf)
        self.K_circle(self.y0,n)
        x,y = (0,self.y0+0.5)
        x2,y2 = self.left_arrow(x,y,self.y0,n)
        self.pulse_text(x2-0.1,y2-0.5,n)
        self.y0 = yf

    def draw_Ku(self,pulse_num,interaction_num):
        """Draws the ket-up (Ku) interaction, which is the rotating term 
            interacting with the ket-side

        Args:
            pulse_num (int) : pulse number
            interaction_num (int) : interaction number
        """
        change = np.array([1,0])
        self.diagram_drawer_manifolds += change
        m = interaction_num
        n = pulse_num
        yf = self.y0 + self.vertical_scale
        self.double_sided(yf)
        self.K_circle(self.y0,n)
        x,y = (-0.5,self.y0)
        x2,y2 = self.right_arrow(x,y,self.y0,n)
        self.pulse_text(x-0.1,y+0.1,n)
        self.y0 = yf

    

class DiagramGenerator(DiagramDrawer):
    """

    Args:
        detection_type (str): default is 'polarization', other options are 'integrated_polarization' or 'fluorescence'

    """
    def __init__(self,*,detection_type = 'polarization'):
        DiagramDrawer.__init__(self)

        # Code will not actually function until the following three empty
        # lists are set by the user
        self.efield_times = [] #initialize empty list of times assoicated with each electric field shape
        self.efield_wavevectors = []

        # Change this, if applicable, to the maximum number of manifolds in the system under study
        self.maximum_manifold = np.inf

        # Change this to a negative number, possibly -np.inf, if the initial state can be de-excited
        self.minimum_manifold = 0

        # Used for automatically generating diagram instructions
        self.wavevector_dict = {'-':('Kd','Bu'),'+':('Ku','Bd')}

        # Used to find resonant-only contributions
        self.instruction_to_manifold_transition = {'Bu':np.array([0,1]),
                                                   'Bd':np.array([0,-1]),
                                                   'Ku':np.array([1,0]),
                                                   'Kd':np.array([-1,0])}
        self.detection_type = detection_type
        
        if 'polarization' in detection_type:
            self.filter_instructions = self.polarization_detection_filter_instructions
        elif detection_type == 'fluorescence':
            self.filter_instructions = self.fluorescence_detection_filter_instructions
        else:
            raise Exception('Detection type not supported')

    def interaction_tup_to_str(self,tup):
        """Converts a tuple, tup = (nr,nc) into a string of +'s and -'s, used
            for converting between different internal methods of representing
            the number of interactions with (counter-)rotating terms for each
            pulse

        Args:
            tup (tuple) : format (nr,nc) where nr (nc) is the number of 
                interactions with the rotating (counter-rotating) term
        Returns:
            str : repeats "+" nr times and "-" nc times
        """
        s = '+'*tup[0] + '-'*tup[1]
        return s

    def set_phase_discrimination(self,interaction_list):
        """Sets the efield wavevectors and pdc (phase discrimination 
            condition), given a list of tuples describing the number of
            interactions with the (counter-)rotating terms of each pulse
        """
        if type(interaction_list[0]) is str:
            new_list = interaction_list
        elif type(interaction_list[0]) is tuple:
            new_list = [self.interaction_tup_to_str(el)
                        for el in interaction_list]

        self.efield_wavevectors = new_list
        self.set_pdc()

        # If pulses and/or phase discrimination are changed, these two
        # attributes must be reset or removed:
        
        try:
            del self.pulse_sequence
            del self.pulse_overlap_array
        except AttributeError:
            pass

    def pdc_arr_to_tup(self,pdc_arr):
        """Converts a pdc represented as a numpy array to a tuple of tuples

        Args:
            pdc_arr (np.ndarray) : pdc as a numpy array
        Returns:
            tuple : pdc as a tuple of tuples
        """
        return tuple(tuple(pdc_arr[i,:].tolist()) for i in range(pdc_arr.shape[0]))

    def pdc_tup_to_arr(self,pdc_tup):
        """Converts a pdc represented as a tuple of tuples to a numpy array

        Args:
            pdc_arr (np.ndarray) : pdc as a tuple of tuples
        Returns:
            tuple : pdc as a numpy array
        """
        return np.array(pdc_tup)
    
    def set_pdc(self):
        """Sets pdc using the attribute "efield_wavevectors"
        """
        num_pulses = len(self.efield_wavevectors)
        pdc = np.zeros((num_pulses,2),dtype='int')
        for i in range(num_pulses):
            for j in range(len(self.efield_wavevectors[i])):
                if self.efield_wavevectors[i][j] == '+':
                    pdc[i,0] += 1
                elif self.efield_wavevectors[i][j] == '-':
                    pdc[i,1] += 1
                else:
                    raise Exception('Could not set phase-discrimination condition')

        self.pulse_orders = pdc.sum(axis=1).tolist()
        self.highest_order = np.sum(pdc)
        self.pdc_tuple = self.pdc_arr_to_tup(pdc)
        self.pdc = pdc
        self.max_pdc = pdc.copy()
        self.set_signal_pdcs([pdc])

    def set_signal_pdcs(self,pdcs):
        """Takes a list of pdcs, converts each pdc to a tuple if necessary, 
            and sets them all to the attribute "signal_pdcs"

        Args:
            pdcs (list) : list of pdcs as np.ndarrays or tuples
        """
        if type(pdcs[0]) is tuple:
            self.signal_pdcs = pdcs
        else:
            self.signal_pdcs = [self.pdc_arr_to_tup(pdc) for pdc in pdcs]
            
        self.all_pdcs = set(self.signal_pdcs)

        return None

    def include_higher_orders(self,pulse_orders):
        """Adds additional higher-order pdcs based upon the baseline pdc 
            stored in attribute "pdc". Additional pdcs are of the same or 
            lower order in each pulse, as specified by pulse_orders
            
        Args:
            pulse_orders (list) : list of integers specifying what order in
                perturbation theory to go to for each pulse
        """
        self.pulse_orders = pulse_orders
        lowest_orders = self.pdc.sum(axis=1)
        highest_orders = np.zeros(lowest_orders.shape,dtype=
                                  lowest_orders.dtype)
        for i in range(len(self.pulse_orders)):
            trial = lowest_orders.copy()
            trial[i] = pulse_orders[i]
            highest_orders[i] = trial.sum()
        self.highest_order = np.max(highest_orders)
        
        self.max_pdc = self.pdc.copy()
        # additional interactions
        add_ints = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)
        for i in range(len(pulse_orders)):
            # additional number of interactions
            add_num_ints = pulse_orders[i] - np.sum(self.pdc[i,:])
            if add_num_ints%2 != 0:
                warnings.warn('Cannot go to order {:0} in pulse {:1}, can only go to order {:2} while satisfying the phase-discrimination condition specified'.format(pulse_orders[i],i,pulse_orders[i]-1))
            # additional interactions are equally divided between the rotating and counter-rotating terms
            add_ints[i,:] = add_num_ints//2
        self.max_pdc += add_ints

        self.set_all_signal_matching_pdcs()

        return None

    def set_all_signal_matching_pdcs(self):
        """Determines all possible higher order pdcs that both match the 
            lowest order pdc, and which are compatible with the specified 
            maximum pulse orders
        """
        signal_pdcs = [self.pdc]
        num_pulses = len(self.pulse_orders)
        for n in range(num_pulses):
            min_order = self.pdc[n,:].sum()
            remainder = self.pulse_orders[n] - min_order
            pdc_add = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)
            pdc_add[n,:] = 1
            new_pdcs = []
            for i in range(remainder // 2):
                for pdc in signal_pdcs:
                    new_pdc = pdc + pdc_add * (i+1)
                    if new_pdc.sum() <= self.highest_order:
                        new_pdcs.append(new_pdc)

            signal_pdcs += new_pdcs

        self.set_signal_pdcs(signal_pdcs)

        return None

    def contract_pdc(self,pdc):
        """Takes a phase-discrimination condition and returns a tuple 
            specifying the wavevector of the resultant signal

        Args:
            pdc (np.ndarray) : input pdc
        Returns:
            tuple : wavevector of resultant signal"""
        sum_rule = np.ones(pdc.shape,dtype=pdc.dtype)
        sum_rule[:,1] = -1
        return tuple(np.sum(sum_rule*pdc,axis=1))

    def save_diagrams_as_text(self,diagrams,*,file_name = 'auto'):
        """Saves diagrams to a text file with each diagram on a separate line

        Args:
            diagrams (list) : list of diagrams
            file_name (str, optional) : filename to use for saving diagrams
        """
        if file_name == 'auto':
            file_name = ''
            for nr,nc in self.pdc_tuple:
                file_name += '{1}_{2}_'.format(nr,nc)
            file_name = file_name[:-1]
            file_name += '.txt'
        with open(file_name,'w') as txtfile:
            txtfile.writelines('['+str(diagrams[0]))
            for diagram in diagrams[1:]:
                txtfile.writelines(',\n '+str(diagram))
            txtfile.writelines(']')

    def polarization_detection_filter_instructions(self,instructions):
        """Filter function to test whether a given diagram ends in a coherence
            between two states that are only one excitation apart

        Args:
            instructions (tuple) : diagram
        Returns:
            bool : True if diagram obeys the polarization detection condition
        """
        rho_manifold = np.array([0,0])
        for ins in instructions:
            rho_manifold += self.instruction_to_manifold_transition[ins]
            if rho_manifold[0] < self.minimum_manifold or rho_manifold[1] < self.minimum_manifold:
                return False
            if rho_manifold[0] > self.maximum_manifold or rho_manifold[1] > self.maximum_manifold:
                return False
        if rho_manifold[0] - rho_manifold[1] == 1:
            return True
        else:
            return False

    def fluorescence_detection_filter_instructions(self,instructions):
        """Filter function to test whether a given diagram ends in an 
            excited-state population

        Args:
            instructions (tuple) : diagram
        Returns:
            bool : True if diagram obeys the fluorescence detection condition
        """
        rho_manifold = np.array([0,0])
        for ins in instructions:
            rho_manifold += self.instruction_to_manifold_transition[ins]
            if rho_manifold[0] < self.minimum_manifold or rho_manifold[1] < self.minimum_manifold:
                return False
            if rho_manifold[0] > self.maximum_manifold or rho_manifold[1] > self.maximum_manifold:
                return False
        if abs(rho_manifold[1]-rho_manifold[1]) == 0 and rho_manifold[1] != 0:
            return True
        else:
            return False

    def RWA_filter_instructions(self,instructions):
        """Filter function to test whether a given diagram obeys the rotating-wave
        approximation (RWA), given that resonance excitations cannot de-excite the
        initial density matrix below attribute "minimum_manifold" and cannot excite
        the initial density matrix above attribute "maximum_manifold"

        Args:
            instructions (tuple) : diagram
        Returns:
            bool : True if diagram obeys the RWA
        """
        rho_manifold = np.array([0,0])
        for ins in instructions:
            rho_manifold += self.instruction_to_manifold_transition[ins]
            if rho_manifold[0] < self.minimum_manifold or rho_manifold[1] < self.minimum_manifold:
                return False
            if rho_manifold[0] > self.maximum_manifold or rho_manifold[1] > self.maximum_manifold:
                return False
        return True

    def instructions_from_permutation(self,perm):
        """Generates a list of diagrams given an input permutation
        
        Args:
            perm (tuple) : tuple of tuples giving a particular ordering of pulse
                interactions
        Returns:
            list : list of Feynman diagrams that obey the input interaction order
                and the detection condition stored in attribute "detection_type"
        """
        f_list = []
        efield_order = []
        for i,k in perm:
            f_list.append(self.wavevector_dict[k])
            efield_order.append(i)
        all_instructions = itertools.product(*f_list)
        filtered_instructions = []
        for ins in all_instructions:
            if self.filter_instructions(ins):
                filtered_instructions.append(tuple(zip(ins,efield_order)))
        return filtered_instructions

    def wavefunction_instructions_from_permutation(self,perm):
        """Generates a list of closed-system diagrams given an input permutation
        
        Args:
            perm (tuple) : tuple of tuples giving a particular ordering of pulse
                interactions
        Returns:
            list : list of Feynman diagrams that obey the input interaction order
                and the detection condition stored in attribute "detection_type"
        """
        rho_instructions = self.instructions_from_permutation(perm)
        psi_instructions = []
        for instructions in rho_instructions:
            new_instructions = self.convert_rho_instructions_to_psi_instructions(instructions)
            if new_instructions in psi_instructions:
                pass
            else:
                psi_instructions.append(new_instructions)

        return psi_instructions

    def convert_rho_instructions_to_psi_instructions(self,instructions):
        """Converts a diagram used for calculating signals using density matrices
            into a diagram that can be used for calculating signals using 
            wavefunctions (sometimes called a loop diagram)
            
        Args:
            instructions (tuple) : a diagram for density matrices
        Returns:
            dict : a diagram for wavefunctions
        """
        psi_instructions = {'ket':[],'bra':[]}
        for key, pulse_num in instructions:
            if key[0] == 'K':
                psi_instructions['ket'].append((key[1],pulse_num))
            elif key[0] == 'B':
                psi_instructions['bra'].append((key[1],pulse_num))
        return psi_instructions

    def convert_rho_instructions_list_to_psi_instructions_list(self,instructions_list):
        """Converts diagrams used for calculating signals using density matrices
            into diagrams that can be used for calculating signals using 
            wavefunctions (sometimes called a loop diagram)
            
        Args:
            instructions (list) : a list of diagrams for density matrices
        Returns:
            list : a list of diagrams for wavefunctions
        """
        psi_instructions_list = []
        for instructions in instructions_list:
            psi_instructions = self.convert_rho_instructions_to_psi_instructions(instructions)
            if psi_instructions not in psi_instructions_list:
                psi_instructions_list.append(psi_instructions)
        return psi_instructions_list

    def relevant_permutations(self,pulse_time_meshes):
        """Determines which permutations of pulse orderings will result in non-zero 
            calculations, given the pulse durations

        Args:
            pulse_time_meshes (list) : list of pulse durations
        Returns:
            list : all permutations of pulses that obey causality
        """
        self.set_ordered_interactions()
        all_permutations = set(itertools.permutations(self.ordered_interactions))
        filtered_permutations = []
            
        for perm in all_permutations:
            remove = False
            for i in range(len(perm)-1):
                indi = perm[i][0]
                ti = pulse_time_meshes[indi]
                for j in range(i+1,len(perm)):
                    indj = perm[j][0]
                    tj = pulse_time_meshes[indj]
                    if ti[0] > tj[-1]:
                        remove = True
                        break
                if remove == True:
                    break
            if not remove:
                filtered_permutations.append(perm)
        return filtered_permutations

    def set_ordered_interactions(self):
        """Sets up a list of the time-ordered interactions, along with associated 
            wavevector ('+' or '-') for each interaction.
        """
        num_pulses = len(self.efield_wavevectors)
        self.ordered_interactions = []
        for i in range(num_pulses):
            ks = self.efield_wavevectors[i]
            for k in ks:
                self.ordered_interactions.append((i,k))

    def check_new_diagram_conditions(self,arrival_times):
        """Check to see if pulse overlaps have changed. If so, a new list of 
            diagrams must be calculated. If not, old list may be reused.
        """
        new = np.array(arrival_times)
        new_pulse_sequence = np.argsort(new)
        new_pulse_overlap_array = np.zeros((len(arrival_times),
                                            len(arrival_times)),
                                           dtype='bool')

        intervals = self.arrival_times_to_pulse_intervals(arrival_times)
        
        for i in range(len(intervals)):
            ti = intervals[i]
            for j in range(i+1,len(intervals)):
                tj = intervals[j]
                if ti[0] >= tj[0] and ti[0] <= tj[-1]:
                    new_pulse_overlap_array[i,j] = True
                elif ti[-1] >= tj[0] and ti[-1] <= tj[-1]:
                    new_pulse_overlap_array[i,j] = True
        try:
            logic_statement = (np.allclose(new_pulse_overlap_array,self.pulse_overlap_array)
                and np.allclose(new_pulse_sequence,self.pulse_sequence))
            if logic_statement:
                calculate_new_diagrams = False
            else:
                calculate_new_diagrams = True
        except AttributeError:
            calculate_new_diagrams = True
        
        self.pulse_sequence = new_pulse_sequence
        self.pulse_overlap_array = new_pulse_overlap_array

        return calculate_new_diagrams

    def arrival_times_to_pulse_intervals(self,arrival_times):
        """Converts the input pulse durations, which are all centered on zero,
            into the intervals of time where each pulse is non-zero, which are
            called pulse intervals.
        
        Args:
            arrival_times (list) : list of pulse arrival times
        Returns:
            list : pulse intervals
        """
        if self.detection_type == 'polarization' or self.detection_type == 'integrated_polarization':
            if len(arrival_times) == len(self.efield_wavevectors) + 1:
                # If the arrival time of the local oscillator was included in the list arrival_times,
                # remove it, it is not relevant to diagram generation
                arrival_times = arrival_times[:-1]
        intervals = [self.efield_times[i] + arrival_times[i] for i in range(len(arrival_times))]

        return intervals

    def set_diagrams(self,arrival_times):
        """Given the input arrival times of each pulse, determine which diagrams
            obey causality, and thus yield a non-zero signal

        Args:
            arrival_times (list) : list of pulse arrival times
        """
        intervals = self.arrival_times_to_pulse_intervals(arrival_times)
        
        efield_permutations = self.relevant_permutations(intervals)
        all_instructions = []
        for perm in efield_permutations:
            all_instructions += self.instructions_from_permutation(perm)
        self.current_diagrams = all_instructions
        

    def get_diagrams(self,arrival_times,*,recalculate=True):
        """Given the input arrival times of each pulse, determine if any pulse
            overlap conditions have changed. If so, calculate a new list of
            relevant diagrams. If not, return previously calculated diagrams.

        Args:
            arrival_times (list) : list of pulse arrival times
        Keyword Args:
            recalculate (bool) : if True, recalculate diagrams
        """
        if recalculate:
            calculate_new_diagrams = True
        else:
            calculate_new_diagrams = self.check_new_diagram_conditions(arrival_times)
        if calculate_new_diagrams:
            self.set_diagrams(arrival_times)
        return self.current_diagrams

    def set_diagram_dict(self,arrival_times):
        """Sets a dictionary of diagrams to calculate, with each key corresponding
            to one pdc stored in attribute "signal_pdcs". Allows for efficient 
            calculation of many different pdcs

        Args:
            arrival_times (list) : list of pulse arrival times
        """
        self.current_diagram_dict = {}
        for pdc in self.signal_pdcs:
            dg = DiagramGenerator(detection_type=self.detection_type)
            dg.efield_times = self.efield_times
            dg.set_phase_discrimination(pdc)
            dg.minimum_manifold = self.minimum_manifold
            dg.maximum_manifold = self.maximum_manifold
            dg.set_diagrams(arrival_times)
            self.current_diagram_dict[pdc] = dg.current_diagrams
        return None

    def get_diagram_dict(self,arrival_times,*,recalculate=False):
        """Given the input arrival times of each pulse, determine if any pulse
            overlap conditions have changed. If so, calculate a new dictionary of
            relevant diagrams. Each key is a pdc from "signal_pdcs".

        Args:
            arrival_times (list) : list of pulse arrival times
        Keyword Args:
            recalculate (bool) : if True, recalculate diagrams
        """
        if recalculate:
            calculate_new_diagrams = True
        else:
            calculate_new_diagrams = self.check_new_diagram_conditions(arrival_times)
        if calculate_new_diagrams:
            self.set_diagram_dict(arrival_times)

        return self.current_diagram_dict

    def get_wavefunction_diagrams(self,arrival_times):
        """Given the input arrival times of each pulse, determine if any pulse
            overlap conditions have changed. If so, calculate a new list of
            relevant wavefunciton diagrams. If not, return previously calculated 
            diagrams.

        Args:
            arrival_times (list) : list of pulse arrival times
        """
        rho_instructions_list = self.get_diagrams(arrival_times)
        wavefunction_instructions = self.convert_rho_instructions_list_to_psi_instructions_list(rho_instructions_list)
        return wavefunction_instructions

    def get_wavefunction_diagram_dict(self,arrival_times):
        """Given the input arrival times of each pulse, determine if any pulse
            overlap conditions have changed. If so, calculate a new dictionary of
            relevant wavefunction diagrams. Each key is a pdc from "signal_pdcs".

        Args:
            arrival_times (list) : list of pulse arrival times
        """
        rho_dict = self.get_diagram_dict(arrival_times)
        psi_dict = {}
        for pdc in self.signal_pdcs:
            psi_dict[pdc] = self.convert_rho_instructions_list_to_psi_instructions_list(rho_dict[pdc])
        return psi_dict

    def get_diagram_final_state(self,diagram):
        """Returns ket and bra manifold indices after all diagram interactions

        Args:
            diagram (list) : list of ordered interactions
        Returns
            list : list with the first entry the ket manifold index, and 
            the second entry the bra manifold index
        """
        rho_manifold = np.array([0,0],dtype='int')
        for ins,pulse_num in diagram:
            rho_manifold += self.instruction_to_manifold_transition[ins]
        return list(rho_manifold)

    def get_diagram_excitation_manifold(self,diagram,*,number_of_interactions=2):
        """Returns the excitation number of the density matrix after the 
            specified number of interactions have occurred. Returns None if the
            density matrix is in a coherence after the given number of
            interactions

        Args:
            diagram (list) : single diagram
        Keyword Args:
            number_of_interactions (int) : check excitation manifold after this 
                number of interactions have occurred
        Returns
            int : excitation number of the density matrix, if it ends in a 
                population (otherwise returns NaN)
        """
        truncated_diagram = diagram[:number_of_interactions]
        rho_manifold = self.get_diagram_final_state(truncated_diagram)
        if rho_manifold[0] == rho_manifold[1]:
            return rho_manifold[0]
        else:
            return np.nan

    def filter_diagrams_by_final_state(self,diagrams,state):
        """Returns all diagrams that end in the specified state

        Args:
            diagrams (list) : list of diagrams to filter
            state (list) : list with the first entry the ket manifold index, 
                and the second entry the bra manifold index
        """
        new_diagrams = []
        for diagram in diagrams:
            diagram_state = self.get_diagram_final_state(diagram)
            if diagram_state == state:
                new_diagrams.append(diagram)
        return new_diagrams

    def filter_diagrams_by_excitation_manifold(self,diagrams,*,manifold=1,number_of_interactions=2):
        """Returns all diagrams that end in the specified excitation manifold or
            excitation number

        Args:
            diagrams (list) : list of diagrams to filter
        Keyword Args:
            manifold (int) : excitation manifold/number to filter for
            number_of_interactions (int) : check excitation manifold after this 
                number of interactions have occurred
        """
        new_diagrams = []
        for diagram in diagrams:
            man = self.get_diagram_excitation_manifold(diagram,number_of_interactions=number_of_interactions)
            if man == manifold:
                new_diagrams.append(diagram)
        return new_diagrams

    def get_diagram_sign(self,diagram):
        """Calculates the overall sign of the input diagram
        
        Args:
            diagram (tuple) : diagram
        """
        if len(diagram) % 2:
            sign = 1j
        else:
            sign = 1
        for ins,pulse_num in diagram:
            if 'K' in ins:
                sign *= -1j
            elif 'B' in ins:
                sign *= 1j
            else:
                raise Exception('diagram is not in proper format')
            
        return sign

    def filter_diagrams_by_sign(self,diagrams,*,sign=1):
        """Filter out only the diagrams that have the specified sign
        
        Args:
            diagrams (list) : list of diagram
        Keyword Args:
            sign (int) : sign of signal (either +1 or -1)
        """
        new_diagrams = []
        for diagram in diagrams:
            diagram_sign = self.get_diagram_sign(diagram)
            if diagram_sign == sign:
                new_diagrams.append(diagram)
        return new_diagrams

    def remove_all_permutations(self,diagrams):
        """Given a list of input diagrams, strip each diagram of duplicate 
            interaction types, and return a list of diagrams where each diagram
            contains only one of each type of interaction. DEPRECATED?
        """
        new_diagrams = []
        diagram_weights = []
        set_list = []
        for diagram in diagrams:
            new_set = set(diagram)
            try:
                ind = set_list.index(new_set)
                diagram_weights[ind] += 1
            except ValueError:
                new_diagrams.append(diagram)
                diagram_weights.append(1)
                set_list.append(new_set)

        return diagram_weights, new_diagrams



import os
from ufss.HLG import PolymerVibrations,DiagonalizeHamiltonian,RedfieldConstructor,SecularRedfieldConstructor,DiagonalizeLiouvillian,OpenPolymerVibrations,convert

class run:

    def __init__(self,folder,*,for_RKE = False,conserve_memory = False,
                 halfway_basis = True):
        self.for_RKE = for_RKE
        self.conserve_memory = conserve_memory
        self.folder = folder
        try:
            convert(folder,halfway_basis=halfway_basis)
        except FileNotFoundError:
            pass

        try:
            pv = PolymerVibrations(os.path.join(folder,'params.yaml'))
        except FileNotFoundError:
            raise Exception('Could not find params.yaml file in specified folder')

        if 'bath' in pv.params.keys():
            self.open = True
            if ('site_bath' in pv.params['bath'].keys() or 
                'vibration_bath' in pv.params['bath'].keys()):
                self.Redfield = True
                try:
                    self.secular = pv.params['bath']['secular']
                except KeyError:
                    self.secular = False
            else:
                self.Redfield = False
                try:
                    self.fdb_flag = pv.params['bath']['force_detailed_balance']
                except KeyError:
                    self.fdb_flag = False
        else:
            self.open = False

        if self.open:
            self.run_open()
        else:
            self.run_closed()

    def run_closed(self):
        if self.for_RKE:
            pass
        else:
            dh = DiagonalizeHamiltonian(self.folder)

    def run_open(self):
        if self.Redfield:
            DiagonalizeHamiltonian(self.folder)
            if self.secular:
                SecularRedfieldConstructor(self.folder,conserve_memory=self.conserve_memory)
            else:
                RedfieldConstructor(self.folder,conserve_memory=self.conserve_memory)
            dl = DiagonalizeLiouvillian(self.folder,conserve_memory=self.conserve_memory,secular=self.secular)
        else:
            opv = OpenPolymerVibrations(os.path.join(self.folder,'params.yaml'),for_RKE=self.for_RKE,force_detailed_balance=self.fdb_flag)



import os
import yaml
from ufss.HLG import run

def make_2LS(folder_name,omega0,tau_deph,kT,secular = True,tau_relax = 'auto',
             conserve_memory = False):
    """Makes a 2LS with the energy separation between the two levels set to 1
    Args:
        folder_name (str) : folder for the calculations
        omega0 (radians / fs) : energy separation of two levels, divided by h bar
        tau_deph (fs) : dephasing time for the system
        kT (radians / fs) : temperature

    Kwargs:
        secular (bool) : use secular Redfield if True, otherwise use full Redfield
        tau_relax (fs) : relaxation time for the excited setate. If 'auto', use tau_deph*1000
    """
    if tau_relax == 'auto':
        tau_relax = tau_deph * 1000

    taud = float(tau_deph * omega0)
    taur = float(tau_relax * omega0)
    kT = float(kT / omega0) # kT must be provided in the units of radians / time, to match omega0

    site_energies = [1] # hbar*omega0 / hbar*omega0

    site_couplings = []

    dipoles = [[1,0,0]]

    trunc_size = 1

    os.makedirs(folder_name,exist_ok=True)

    site_bath = {'dephasing_rate':1/taud, # additonal "pure" dephasing
              'relaxation_rate':0,
               'temperature':kT,
                'spectrum_type':'white-noise'}

    vibration_bath = site_bath # use same bath for sites and vibrations

    relax_bath = {'dephasing_rate':0, # additonal "pure" dephasing
              'relaxation_rate':1/taur,
               'temperature':kT,
                'spectrum_type':'white-noise'}

    Redfield_bath = {'secular':secular,
        'site_bath':site_bath,
        'vibration_bath':vibration_bath,
        'site_internal_conversion_bath':relax_bath}

    vibrations = []

    params = {
        'site_energies':site_energies,
        'site_couplings':site_couplings,
        'dipoles':dipoles,
        'truncation_size':trunc_size,
        'vibrations':vibrations,
        'bath':Redfield_bath,
        'maximum_manifold':3}

    with open(os.path.join(folder_name,'simple_params.yaml'),'w+') as new_file:
        yaml.dump(params,new_file)

    run(folder_name,conserve_memory=conserve_memory)

def make_3LS(folder_name,omega0,tau_deph,kT,secular = True,tau_relax = 'auto',
             conserve_memory = False,omega21 = 'auto',dipole12 = 'auto',
             tau_deph2 = 'auto',tau_relax2 = 'auto'):
    """Makes a 2LS with the energy separation between the two levels set to 1
    Args:
        folder_name (str) : folder for the calculations
        omega0 (radians / fs) : energy separation of two levels, divided by h bar
        tau_deph (fs) : dephasing time for the system
        kT (radians / fs) : temperature

    Kwargs:
        secular (bool) : use secular Redfield if True, otherwise use full Redfield
        tau_relax (fs) : relaxation time for the excited setate. If 'auto', use tau_deph*1000
    """
    if tau_relax == 'auto':
        tau_relax = tau_deph * 1000
    if omega21 == 'auto':
        omega21 = 2 * omega0
    if dipole12 == 'auto':
        dipole12 = [1,0,0]
    if tau_deph2 == 'auto':
        tau_deph2 = tau_deph
    if tau_relax2 == 'auto':
        tau_relax2 = tau_relax

    taud = float(tau_deph * omega0)
    taud2 = float(tau_deph2 * omega0)
    taur = float(tau_relax * omega0)
    taur2 = float(tau_relax2 * omega0)
    kT = float(kT / omega0) # kT must be provided in the units of radians / time, to match omega0
    site_energies = [[1,float(omega21/omega0)],]

    site_couplings = [[],[],[]]

    dipole01 = [1,0,0]
    dipole02 = [0,0,0]

    dipoles = [[dipole01,dipole12,dipole02],]

    trunc_size = 1

    os.makedirs(folder_name,exist_ok=True)

    site_bath = {'dephasing_rate':1/taud, # additonal "pure" dephasing
              'relaxation_rate':0,
               'temperature':kT,
                'spectrum_type':'white-noise'}
    
    site_bath2 = {'dephasing_rate':1/taud2, # additonal "pure" dephasing
              'relaxation_rate':0,
               'temperature':kT,
                'spectrum_type':'white-noise'}

    vibration_bath = site_bath # use same bath for sites and vibrations

    relax_bath = {'dephasing_rate':0, # additonal "pure" dephasing
              'relaxation_rate':1/taur,
               'temperature':kT,
                'spectrum_type':'white-noise'}
    
    relax_bath2 = {'dephasing_rate':0, # additonal "pure" dephasing
              'relaxation_rate':1/taur2,
               'temperature':kT,
                'spectrum_type':'white-noise'}

    Redfield_bath = {'secular':secular,
        'site_bath':site_bath,
        'site_bath2':site_bath2,
        'vibration_bath':vibration_bath,
        'site_internal_conversion_bath':relax_bath,
        'site_internal_conversion_bath21':relax_bath2}

    vibrations = []

    params = {
        'site_energies':site_energies,
        'site_couplings':site_couplings,
        'dipoles':dipoles,
        'truncation_size':trunc_size,
        'vibrations':vibrations,
        'bath':Redfield_bath,
        'maximum_manifold':3}

    with open(os.path.join(folder_name,'simple_params.yaml'),'w+') as new_file:
        yaml.dump(params,new_file)

    run(folder_name,conserve_memory=conserve_memory)
